local _G = _G
local Cartographer3 = _G.Cartographer3
local L = Cartographer3.L

Cartographer3.Options = {}

function Cartographer3.Options.MakeSuboptionsPanel(name, controlCreationFunc)
	local panel = CreateFrame("Frame", "Cartographer3_Panel_" .. name)
	
	panel.name = name
	panel.controls = {}
	if name ~= "Cartographer3" then
		panel.parent = "Cartographer3"
	end
	
	function panel:okay()
		for control in pairs(self.controls) do
			control.oldValue = control.value
			if control.okayFunc then
				control:okayFunc()
			end
		end
	end

	function panel:cancel()
		for control in pairs(self.controls) do
			control:SetValue(control.oldValue)
			if control.cancelFunc then
				control:cancelFunc()
			end
		end
	end

	function panel:default()
		for control in pairs(self.controls) do
			control:SetValue(control.default)
			if control.defaultFunc then
				control:defaultFunc()
			end
		end
	end
	
	InterfaceOptions_AddCategory(panel)
	
	panel:SetScript("OnShow", function(self)
		self:SetScript("OnShow", nil)
		controlCreationFunc(self)
		controlCreationFunc = nil
	end)
end

function Cartographer3.Options.MakeTitleTextAndSubText(self, titleText, subTextText)
	local title = self:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetText(titleText)
	title:SetJustifyH("LEFT")
	title:SetJustifyV("TOP")
	title:SetPoint("TOPLEFT", 16, -16)
	
	local subText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	subText:SetText(subTextText)
	subText:SetNonSpaceWrap(true)
	subText:SetJustifyH("LEFT")
	subText:SetJustifyV("TOP")
	subText:SetPoint("TOPLEFT", title, "BOTTOMLEFT", 0, -8)
	subText:SetPoint("RIGHT", -32, 0)
	
	return title, subText
end

function Cartographer3.Options.MakeSlider(self, text, description, lowText, farText, minValue, maxValue, step, default, current, changeFunc, currentTextFunc)
	local name
	local i = 0
	repeat
		i = i + 1
		name = self:GetName() .. "_Slider" .. i
	until not _G[name]
	local slider = CreateFrame("Slider", name, self, "InterfaceOptionsSliderTemplate")
	self.controls[slider] = true
	_G[slider:GetName() .. "Text"]:SetText(text)
	slider.tooltipText = description
	_G[slider:GetName() .. "Text"]:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	_G[slider:GetName() .. "Low"]:SetText(lowText)
	_G[slider:GetName() .. "Low"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	_G[slider:GetName() .. "High"]:SetText(farText)
	_G[slider:GetName() .. "High"]:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b)
	
	local currentText
	if currentTextFunc then
		currentText = slider:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
		currentText:SetPoint("TOP", slider, "CENTER", 0, -8)
		currentText:SetText(currentTextFunc(current))
	end
	
	slider.default = default
	slider:SetMinMaxValues(minValue, maxValue)
	slider:SetValueStep(step)
	slider.oldValue = current
	slider.value = current
	slider:SetValue(current)
	local oldSetValue = slider.SetValue
	function slider:SetValue(value)
		oldSetValue(self, value)
		self.value = value
		changeFunc(value)
		if currentText then
			currentText:SetText(currentTextFunc(value))
		end
	end
	slider:SetScript("OnValueChanged", function(self)
		self.value = self:GetValue()
		self:SetValue(self:GetValue())
	end)
	return slider
end

function Cartographer3.Options.MakeDropDown(self, text, description, values, default, current, changeFunc)
	local name
	local i = 0
	repeat
		i = i + 1
		name = self:GetName() .. "_DropDown" .. i
	until not _G[name]
	
	local dropDown = CreateFrame("Frame", name, self, "UIDropDownMenuTemplate")
	self.controls[dropDown] = true
	local label = dropDown:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
	label:SetText(text)
	dropDown.tooltipText = description
	label:SetPoint("BOTTOMLEFT", dropDown, "TOPLEFT", 16, 3)
	UIDropDownMenu_Initialize(dropDown, function()
		for i = 1, #values, 2 do
			local value = values[i]
			local text = values[i+1]

			local info = UIDropDownMenu_CreateInfo()
			info.text = text
			info.value = value
			info.checked = dropDown.value == value
			info.func = dropDown.SetValue
			info.arg1 = dropDown
			info.arg2 = value
			UIDropDownMenu_AddButton(info)
		end
	end)
	UIDropDownMenu_SetWidth(90, dropDown)
	UIDropDownMenu_SetSelectedValue(dropDown, Cartographer3.db.strata)
	dropDown.default = default
	dropDown.value = current
	dropDown.oldValue = current
	function dropDown:SetValue(value)
		self.value = value
		changeFunc(value)
		UIDropDownMenu_SetSelectedValue(dropDown, value)
	end
	dropDown:EnableMouse(true)
	dropDown:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_TOPRIGHT")
		GameTooltip:SetText(self.tooltipText, nil, nil, nil, nil, 1)
	end)
	dropDown:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	return dropDown
end

function Cartographer3.Options.MakeButton(self, text, description, clickFunc)
	local name
	local i = 0
	repeat
		i = i + 1
		name = self:GetName() .. "_Button" .. i
	until not _G[name]
	
	local button = CreateFrame("Button", name, self, "UIPanelButtonTemplate2")
	self.controls[button] = true
	button:SetText(text)
	button.tooltipText = description
	button:SetWidth(120)
	button:SetHeight(22)
	function button:SetValue()
	end
	button:SetScript("OnClick", function(self)
		clickFunc()
	end)
	button:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:SetText(self.tooltipText, nil, nil, nil, nil, 1)
	end)
	button:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	return button
end

function Cartographer3.Options.MakeToggle(self, text, description, default, current, changeFunc)
	local name
	local i = 0
	repeat
		i = i + 1
		name = self:GetName() .. "_Toggle" .. i
	until not _G[name]
	
	local toggle = CreateFrame("CheckButton", name, self, "InterfaceOptionsCheckButtonTemplate")
	self.controls[toggle] = true
	_G[toggle:GetName() .. "Text"]:SetText(text)
	toggle.default = default
	toggle.value = current
	toggle.oldValue = current
	function toggle:SetValue(value)
		changeFunc(not not value)
		self.value = value
		self:SetChecked(value)
	end	
	toggle:SetScript("OnClick", function(self)
		self:SetValue(not not self:GetChecked())
	end)	
	toggle:SetChecked(current)
	toggle:SetScript("OnEnter", function(this)
		GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
		GameTooltip:SetText(description, nil, nil, nil, nil, 1)
	end)
	toggle:SetScript("OnLeave", function(this)
		GameTooltip:Hide()
	end)
	return toggle
end

function Cartographer3.Options.MakeColorPicker(self, nameText, description, hasAlpha, defaultR, defaultG, defaultB, defaultA, currentR, currentG, currentB, currentA, changeFunc)
	local name
	local i = 0
	repeat
		i = i + 1
		name = self:GetName() .. "_ColorPicker" .. i
	until not _G[name]
	
	if not hasAlpha then
		defaultA = 1
		currentA = 1
	end
	
	local button = CreateFrame("Button", name, self)
	self.controls[button] = true
	button:SetHitRectInsets(0, -100, 0, 0)
	local text = button:CreateFontString(nil, "ARTWORK", "GameFontHighlight")
	text:SetText(nameText)
	text:SetPoint("LEFT", button, "RIGHT", 0, 1)
	local color = button:CreateTexture(nil, "ARTWORK")
	button.color = color
	color:SetTexture(currentR, currentG, currentB, currentA)
	local background = button:CreateTexture(nil, "BORDER")
	background:SetTexture([=[Tileset\Generic\Checkers]=])
	background:SetTexCoord(0, 0.5, 0, 0.5)
	local border = button:CreateTexture(nil, "BACKGROUND")
	border:SetTexture([=[Interface\ChatFrame\ChatFrameColorSwatch]=])
	button:SetWidth(26)
	button:SetHeight(26)
	background:SetPoint("CENTER")
	background:SetWidth(14)
	background:SetHeight(14)
	color:SetPoint("CENTER")
	color:SetWidth(14)
	color:SetHeight(14)
	border:SetAllPoints(button)
	button.default = { defaultR, defaultG, defaultB }
	button.oldValue = { currentR, currentG, currentB }
	if hasAlpha then
		button.default[4] = defaultA
		button.oldValue[4] = currentA
	end
	button.value = button.oldValue
	local info
	local function func(r, g, b, a)
		if not hasAlpha then
			a = 1
		end
		info.r = r
		info.g = g
		info.b = b
		if hasAlpha then
			info.opacity = a
		end
		color:SetTexture(r, g, b, a)
		if button.value == button.oldValue then
			button.value = {}
		end
		button.value[1] = r
		button.value[2] = g
		button.value[3] = b
		if hasAlpha then
			button.value[4] = a
		end
		info.r = r
		info.g = g
		info.b = b
		if hasAlpha then
			info.opacity = 1 - a
		end
		if hasAlpha then
			changeFunc(r, g, b, a)
		else
			changeFunc(r, g, b)
		end
	end
	function button:SetValue(value)
		func(unpack(value))
	end
	local function swatchFunc()
		local r, g, b = ColorPickerFrame:GetColorRGB()
		local opacity = 1 - OpacitySliderFrame:GetValue()
		
		func(r, g, b, opacity)
	end
	info = {
		swatchFunc = swatchFunc,
		hasOpacity = hasAlpha,
		opacityFunc = hasAlpha and swatchFunc or nil,
		r = currentR,
		g = currentB,
		b = currentG,
		opacity = hasAlpha and 1 - currentA or nil,
		cancelFunc = function(...)
			local previousValues = ColorPickerFrame.previousValues
			local r, g, b, opacity = previousValues.r, previousValues.g, previousValues.b, hasAlpha and 1 - previousValues.opacity or 1
			
			func(r, g, b, opacity)
		end,
	}
	button:SetScript("OnClick", function(this)
		OpenColorPicker(info)
	end)
	button:SetScript("OnEnter", function(this)
		GameTooltip:SetOwner(this, "ANCHOR_RIGHT")
		GameTooltip:SetText(description, nil, nil, nil, nil, 1)
	end)
	button:SetScript("OnLeave", function(this)
		GameTooltip:Hide()
	end)
	button:RegisterForClicks("LeftButtonUp")
	return button
end

Cartographer3.Options.MakeSuboptionsPanel("Cartographer3", function(self)
	local title, subText = Cartographer3.Options.MakeTitleTextAndSubText(self, "Cartographer3", L["These options allow you to configure Cartographer3"])
	
	local versionText = self:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmall")
	local format, height, website
	if Cartographer3.versionType == "Development" then
		format = "You are running %s from the source code repository. This is for |cffff7f7fdevelopers only|r. It is recommended in nearly all cases to instead use an alpha version from %s" -- no need for localization
		height = 40
		website = "http://wow.curseforge.com/projects/cartographer3/"
	elseif Cartographer3.versionType == "Alpha" then
		format = L["Version %s: This is for alpha testers wanting to try out new features and have possibly bad bugfixes |cffff7f7fbefore they are ready|r. It is recommended to instead use a beta or release version from %s"]
		height = 40
		website = "curse.com"
	elseif Cartographer3.versionType == "Beta" then
		format = L["Version %s: This is a beta version for those who want to test new features and bugfixes before they are potentially ready. It is recommended to use a release version from %s"]
		height = 40
		website = "curse.com"
	elseif Cartographer3.versionType == "Release" then
		format = L["Version %s. There may be more up-to-date versions on %s"]
		height = 20
		website = "curse.com"
	end
	versionText:SetText(format:format(Cartographer3.version, website))
	versionText:SetNonSpaceWrap(true)
	versionText:SetJustifyH("LEFT")
	versionText:SetJustifyV("TOP")
	versionText:SetPoint("TOPLEFT", subText, "BOTTOMLEFT", 0, -8)
	versionText:SetPoint("RIGHT", -32, 0)
	versionText:SetHeight(height)
	
	local openReportIssueBox, openDonateBox
	do
		local url
		local function makeURLFrame()
			makeURLFrame = nil
			local function bumpFrameLevels(frame, amount)
				frame:SetFrameLevel(frame:GetFrameLevel()+amount)
				local children = newList(frame:GetChildren())
				for _,v in ipairs(children) do
					bumpFrameLevels(v, amount)
				end
				children = del(children)
			end
			-- some code borrowed from Prat here
			StaticPopupDialogs["CARTOGRAPHER3_SHOW_URL"] = {
				text = not IsMacClient() and L["Press Ctrl-C to copy, then Alt-Tab out of the game, open your favorite web browser, and paste the link into the address bar."] or L["Press Cmd-C to copy, then Cmd-Tab out of the game, open your favorite web browser, and paste the link into the address bar."],
				button2 = ACCEPT,
				hasEditBox = 1,
				hasWideEditBox = 1,
				showAlert = 1, -- HACK : it's the only way I found to make de StaticPopup have sufficient width to show WideEditBox :(

				OnShow = function()
					local editBox = _G[this:GetName() .. "WideEditBox"]
					editBox:SetText(url)
					editBox:SetFocus()
					editBox:HighlightText(0)
					editBox:SetScript("OnTextChanged", function() StaticPopup_EditBoxOnTextChanged() end)

					local button = _G[this:GetName() .. "Button2"]
					button:ClearAllPoints()
					button:SetWidth(200)
					button:SetPoint("CENTER", editBox, "CENTER", 0, -30)

					_G[this:GetName() .. "AlertIcon"]:Hide()  -- HACK : we hide the false AlertIcon
					this:SetFrameStrata("FULLSCREEN_DIALOG")
					this:SetFrameLevel(this:GetFrameLevel()+30)
				end,
				OnHide = function()
					local editBox = _G[this:GetName() .. "WideEditBox"]
					editBox:SetScript("OnTextChanged", nil)
					this:SetFrameStrata("DIALOG")
					this:SetFrameLevel(this:GetFrameLevel()-30)
				end,
				OnAccept = function() end,
				OnCancel = function() end,
				EditBoxOnEscapePressed = function() this:GetParent():Hide() end,
				EditBoxOnTextChanged = function()
					this:SetText(url)
					this:SetFocus()
					this:HighlightText(0)
				end,
				timeout = 0,
				whileDead = 1,
				hideOnEscape = 1
			}
		end
		
		function openReportIssueBox()
			if makeURLFrame then
				makeURLFrame()
			end
		
			url = "http://wow.curseforge.com/projects/cartographer3/tickets/"

			StaticPopup_Show("CARTOGRAPHER3_SHOW_URL")
		end
		
		function openDonateBox()
			if makeURLFrame then
				makeURLFrame()
			end
			
			url = "https://www.paypal.com/cgi-bin/webscr?cmd=_xclick&business=ckknight%40gmail%2ecom&item_name=Cartographer3"

			StaticPopup_Show("CARTOGRAPHER3_SHOW_URL")
		end
	end
	
	local reportButton = Cartographer3.Options.MakeButton(self,
		L["Report Issue"],
		L["Report a defect or a possible enhancement you see with Cartographer3."],
		openReportIssueBox)
	reportButton:SetPoint("TOPLEFT", versionText, "BOTTOMLEFT", 0, -16)
	
	local donateButton = Cartographer3.Options.MakeButton(self,
		L["Donate"],
		L["Give a much-needed donation to ckknight, the author of Cartographer3."],
		openDonateBox)
	donateButton:SetPoint("TOPRIGHT", versionText, "BOTTOMRIGHT", 0, -16)
	
	local zoomToMinimapTextureSlider = Cartographer3.Options.MakeSlider(self,
		L["Texture Change Zoom"],
		L["The amount to zoom in before the map changes from minimap-style textures to artwork.\n\nNote: Setting this too far may cause severe framerate issues."],
		L["Far"],
		L["Near"],
		0, 120,
		1,
		Cartographer3.Data.DEFAULT_ZOOM_TO_MINIMAP_TEXTURE,
		Cartographer3.db.zoomToMinimapTexture,
		function(value)
			Cartographer3.db.zoomToMinimapTexture = value
		end)
	zoomToMinimapTextureSlider:SetPoint("TOPLEFT", reportButton, "BOTTOMLEFT", 0, -32)
	
	local strataDropDown = Cartographer3.Options.MakeDropDown(self,
		L["Strata"],
		L["What level on the interface the map will show up on. 'Fullscreen' is the highest possible level and 'Background' is the lowest."],
		{
			"FULLSCREEN", L["Fullscreen"],
			"DIALOG", L["Dialog"],
			"HIGH", L["High"],
			"MEDIUM", L["Medium"],
			"LOW", L["Low"],
			"BACKGROUND", L["Background"]
		},
		"FULLSCREEN",
		Cartographer3.db.strata,
		function(value)
			Cartographer3.SetFrameStrata(value)
		end)
	strataDropDown:SetPoint("TOPRIGHT", donateButton, "BOTTOMRIGHT", 13, -28)
	
	local closeWithEscapeToggle = Cartographer3.Options.MakeToggle(self,
		L["Close map with escape key"],
		L["Set whether pressing the escape key will close the map or not."],
		true,
		Cartographer3.db.closeWithEscape,
		function(value)
			Cartographer3.db.closeWithEscape = value
		end)
	closeWithEscapeToggle:SetPoint("TOPLEFT", zoomToMinimapTextureSlider, "BOTTOMLEFT", 0, -24)
	
	local opacitySlider = Cartographer3.Options.MakeSlider(self,
		L["Opacity"],
		L["The opacity of the Cartographer3 map"],
		"10%", "100%",
		0.1, 1,
		0.05,
		1,
		Cartographer3.db.opacity,
		function(value)
			Cartographer3.SetOpacity(value)
		end,
		function(value)
			return ("%.0f%%"):format(value * 100)
		end)
	opacitySlider:SetPoint("TOPLEFT", closeWithEscapeToggle, "BOTTOMLEFT", 0, -24)
	
	local unexploredColorPicker = Cartographer3.Options.MakeColorPicker(self,
		L["Unexplored color"],
		L["Set the color of unexplored areas on the Cartographer3 map. White shows everything and black obscures everything."],
		true,
		1, 1, 1, 1,
		Cartographer3.db.unexploredColor[1], Cartographer3.db.unexploredColor[2], Cartographer3.db.unexploredColor[3], Cartographer3.db.unexploredColor[4],
		function(r, g, b, a)
			Cartographer3.SetUnexploredColor(r, g, b, a)
		end)
	unexploredColorPicker:SetPoint("TOPLEFT", opacitySlider, "BOTTOMLEFT", 0, -32)
	
	local alwaysShowBorderToggle = Cartographer3.Options.MakeToggle(self,
		L["Always show border"],
		L["Set whether the border should always show or whether it should fade out when not hovering over it."],
		false,
		Cartographer3.db.alwaysShowBorder,
		function(value)
			Cartographer3.db.alwaysShowBorder = value
		end)
	alwaysShowBorderToggle:SetPoint("TOPLEFT", unexploredColorPicker, "BOTTOMLEFT", 0, -16)
end)

Cartographer3.Options.MakeSuboptionsPanel(L["POIs"], function(self)
	local title, subText = Cartographer3.Options.MakeTitleTextAndSubText(self, L["POIs"], L["These are options to change how points of interest (POIs) look"])
	
	local last = subText
	
	local t = {}
	for id in pairs(Cartographer3.Data.poiTypes) do
		t[#t+1] = id
	end
	table.sort(t)
	for _, id in ipairs(t) do
		local localizedName = Cartographer3.Data.poiTypes[id]
		local exampleFrame = Cartographer3.Data.poiTypeExampleFrames[id]()
		Cartographer3.Data.poiTypeExampleFrames[id] = nil
		
		
		local function resize(value)
			local scale = Cartographer3.Data.poiTypeScales[id]
			exampleFrame:SetWidth(scale * 20 * value)
			exampleFrame:SetHeight(scale * 20 * value)
		end
		
		local sizeSlider = Cartographer3.Options.MakeSlider(self,
			localizedName,
			L["The size that %s appears on the map"]:format(localizedName),
			"25%", "400%",
			0.25, 4,
			0.05,
			1,
			Cartographer3.db.pois[id],
			function(value)
				Cartographer3.db.pois[id] = value
				for poi in pairs(Cartographer3.Data.shownPOIs) do
					poi:Resize()
				end
				resize(value)
			end,
			function(value)
				return ("%.0f%%"):format(value * 100)
			end)
		
		sizeSlider:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -24)
		
		exampleFrame:SetParent(self)
		exampleFrame:SetPoint("CENTER", sizeSlider, "RIGHT", 100, 0)
		
		resize(Cartographer3.db.pois[id])
		
		last = sizeSlider
	end
end)

Cartographer3.Options.MakeSuboptionsPanel(L["Advanced"], function(self)
	local title, subText = Cartographer3.Options.MakeTitleTextAndSubText(self, L["Advanced"], L["These are advanced options for Cartographer3"])
	
	local hijackWorldMapToggle = Cartographer3.Options.MakeToggle(self,
		L["Separate Cartographer3 from the World Map"],
		L["Set whether to separate Cartographer3 from the World Map.\n\nNote: This will prevent most third-party addons from working properly.\n\nYou can set the keybinding to open and close Cartographer3 in the Keybindings menu.\n\nChanging this initiates a |cffff0000ReloadUI|r"],
		false,
		not Cartographer3.db.hijackWorldMap,
		function(value)
			Cartographer3.db.hijackWorldMap = not value
		end)
	function hijackWorldMapToggle:okayFunc()
		if Cartographer3.db.hijackWorldMap ~= Cartographer3.hijackingWorldMap then
			ReloadUI()
		end
	end	
	hijackWorldMapToggle.cancelFunc = hijackWorldMapToggle.okayFunc
	hijackWorldMapToggle.defaultFunc = hijackWorldMapToggle.okayFunc
	hijackWorldMapToggle:SetPoint("TOPLEFT", subText, "BOTTOMLEFT", 0, -16)
end)

Cartographer3.Options.MakeSuboptionsPanel(L["Modules"], function(self)
	local title, subText = Cartographer3.Options.MakeTitleTextAndSubText(self, L["Modules"], L["Enable or disable the modules you want"])
	
	local modules = {}
	for k in pairs(Cartographer3.modules) do
		modules[#modules+1] = k
	end
	for k in pairs(Cartographer3.potentialModules) do
		modules[#modules+1] = k
	end
	table.sort(modules, function(alpha, bravo)
		local alpha_name = Cartographer3.modules[alpha] and Cartographer3.modules[alpha].name or Cartographer3.potentialModules[alpha][1]
		local bravo_name = Cartographer3.modules[bravo] and Cartographer3.modules[bravo].name or Cartographer3.potentialModules[bravo][1]
		
		return alpha_name < bravo_name
	end)
	
	local last = subText
	
	for _, name in ipairs(modules) do
		local enableStatusToggle = Cartographer3.Options.MakeToggle(self,
			Cartographer3.modules[name] and Cartographer3.modules[name].name or Cartographer3.potentialModules[name][1],
			Cartographer3.modules[name] and Cartographer3.modules[name].desc or Cartographer3.potentialModules[name][2],
			true,
			not Cartographer3.db[name] or not Cartographer3.db[name].disabled,
			function(value)
				if not Cartographer3.db[name] then
					Cartographer3.db[name] = {}
				end
				local disabled = not Cartographer3.db[name].disabled or nil
				Cartographer3.db[name].disabled = disabled
				if disabled then
					if Cartographer3.modules[name] and Cartographer3.modules[name].Disable then
						Cartographer3.modules[name].Disable()
					end
				else
					if Cartographer3.modules[name] then
						if Cartographer3.modules[name].Enable then
							Cartographer3.modules[name].Enable()
						end
					else
						LoadAddOn("Cartographer3_" .. name)
					end
				end
			end)
		
		enableStatusToggle:SetPoint("TOPLEFT", last, "BOTTOMLEFT", 0, -24)
		
		last = enableStatusToggle
	end
end)